<?php
/**
 * Helper functions for the URL token module.
 */


/**
 * Validate a token.
 * By default, this also marks the token as used (incrementing the usage count,
 * and removing the token if this use passes the usage limit).
 *
 * The recommended way to validate a token is directly in a URL, by
 * implementing hook_menu() with a wildcard loader in the path, for example:
 *
 * $item['/foo/bar/%urltoken'] = array(
 *   'title' => 'foo',
 *   // Provide your own page callback like any other hook_menu implementation.
 *   'page callback' => 'foo_serve_page',
 *   // REQUIRED: Pass the name of the module as a load argument.
 *   'load arguments' => array('foo'),
 *   // The %urltoken wildcard handler will make sure the token is valid.
 *   // You can add additional access controls here if desired (for example,
 *   // you might decide that users also have to be logged in before they can
 *   // use the token).
 *   'access callback' => TRUE,
 * );
 *
 * @see urltoken_load()
 *
 * @param String $token
 * The token value.
 * @param String $module
 * Name of the module that requested the token.
 * @param optional String $delta
 * Unique ID allowing a module to provide different types of token, and
 * differentiate between types.
 * @param optional Boolean $mark_token_as_used
 * Set to FALSE to validate the token, without marking the token as used (and
 * therefore increasing the usage count, where tokens are limited to a set
 * number of uses).
 *
 * @return Boolean
 * TRUE if the token is valid.
 */
function urltoken_validate_token($token, $module, $delta = '', $mark_token_as_used = TRUE) {
  // Track all checked tokens in a static variable, so that multiple calls
  // within the same request do not create multiple usage marks.
  $checked_tokens = &drupal_static(__FUNCTION__);
  if (is_null($checked_tokens)) {
    $checked_tokens = array();
  }

  if (!array_key_exists($token, $checked_tokens)) {
    $checked_tokens[$token] = FALSE;

    // Look up validity limits and usage.
    $result = db_select('urltoken', 'u')
      ->fields('u', array('validity_time_limit', 'validity_count_limit', 'usage_count', 'expires'))
      ->condition('token', $token)
      ->condition('module', $module)
      ->condition('delta', $delta)
      ->execute()
      ->fetchObject();

    // Token not found, therefore invalid.
    if (!$result) {
      return FALSE;
    }

    // Check that the time-limit hasn't expired.
    if ($result->validity_time_limit && $result->expires < REQUEST_TIME) {
      urltoken_remove_token($token);
      return FALSE;
    }

    // Increment the usage count (if the token is to be marked as used).
    if ($mark_token_as_used) {
      $result->usage_count++;

      // If this was the last available use of the token, remove the token
      if ($result->validity_count_limit && $result->usage_count >= $result->validity_count_limit) {
        urltoken_remove_token($token);
        return FALSE;
      }
      // Otherwise, simply write the new usage_count to the db.
      // Track the usage-count even if there is no count validity limit.
      else {
        db_update('urltoken')
          ->fields(array(
            'usage_count' => $result->usage_count,
          ))
          ->condition('token', $token)
          ->execute();
      }
    }

    // Success; the token is valid.
    $checked_tokens[$token] = $token;    
  }

  return $checked_tokens[$token];
}


/**
 * Create a token.
 * By default, tokens are valid for 1 hour.
 *
 * @param String $module
 * Name of the module requesting a token.
 * @param optional String $delta
 * If a module needs more than 1 type of token, provide a delta to distinguish
 * between each type.
 * @param optional Int $time_limit
 * Limit the validity of the token to a fixed period of time: defaults to 1
 * hour. Set to 0 to remove any time limit.
 * @param optional Int $count_limit
 * Number of times the token can be used. Defaults to 0 (no restrictions).
 *
 * @return String
 * The authentication token.
 */
function urltoken_create_token($module, $delta = '', $time_limit = URL_TOKEN_DEFAULT_USE_TIME_LIMIT, $count_limit = 0) {
  $token = urltoken_generate_token();

  // Expiry-time is set to NULL if there is no time-limit on validity.
  db_update('urltoken')
    ->fields(array(
      'module'               => $module,
      'delta'                => $delta,
      'validity_time_limit'  => $time_limit,
      'validity_count_limit' => $count_limit,
      'usage_count'          => 0,
      'created'              => REQUEST_TIME,
      'expires'              => ($time_limit) > 0 ? REQUEST_TIME + $time_limit : NULL,
    ))
    ->condition('token', $token)
    ->execute();

  return $token;
}

/**
 * Reset all limits on a single token.
 * Resets the usage count and starts the time limit again.
 *
 * @param String $token
 * The token.
 */
function urltoken_reset_token_limits($token) {
  // Look up the current time limit.
  $current_time_limit = db_select('urltoken', 'u')
    ->fields('u', 'validity_time_limit')
    ->condition('token', $token)
    ->execute()
    ->fetchField();

  db_update('urltoken')
    ->fields(array(
      'usage_count' => 0,
      'created' => REQUEST_TIME,
      'expires' => ($current_time_limit) > 0 ? REQUEST_TIME + $current_time_limit : NULL,
    ))
    ->condition('token', $token)
    ->execute();
}

/**
 * Remove a single token.
 *
 * @param String $token
 * The token.
 */
function urltoken_remove_token($token) {
  db_delete('urltoken')
    ->condition('token', $token)
    ->execute();
}

/**
 * Remove all the tokens attached to a module (or a module-delta combination).
 *
 * @param String $module
 * Name of the module that requested the token(s).
 * @param optional String $delta
 * If the module uses different types of token, you can use a delta to
 * differentiate between types.
 */
function urltoken_remove_all_tokens($module, $delta = NULL) {
  $query = db_delete('urltoken')
    ->condition('module', $module);

  // If a delta is provided, add a clause for the delta.
  if (!is_null($delta)) {
    $query->condition('delta', $delta);
  }

  $query->execute();
}


/**
 * Generate a token.
 * New tokens are checked against the list of current tokens, to guarantee
 * uniqueness. When a token is created, it's immediately added to the DB.
 *
 * @return String
 * A unique token.
 */
function urltoken_generate_token() {
  while (TRUE) {
    try {
      $token = user_password(variable_get('urltoken_length', URLTOKEN_DEFAULT_TOKEN_LENGTH));

      // Attempt to insert it, duplicate keys will trigger an exception which
      // is caught.
      db_insert('urltoken')
        ->fields(array(
          'token' => $token,
        ))
        ->execute();

      return $token;
    }
    // Token is a primary key, so duplicate keys will trigger an exception.
    catch (Exception $e) {
      // Nothing to handle, just loop, so that a new token is generated.
    }
  }
}
